import warnings

import astropy.units as u
import numpy as np
from astropy.coordinates import SkyOffsetFrame
from astropy.time import Time

__all__ = ['Chord']


class Chord:
    """Defines an Occultation Chord.

    Attributes
    ----------
    name : `str`
        The name of the Chord.

    observer : `sora.Observer`
        The site of observation.

    lightcurve : `sora.LightCurve`
        The lightcurve observed.

    """

    def __init__(self, *, name, observer, lightcurve):

        from sora.lightcurve import LightCurve
        from sora.observer import Observer, Spacecraft

        if not isinstance(observer, (Observer, Spacecraft)):
            raise ValueError('obs must be an Observer object')
        if not isinstance(lightcurve, LightCurve):
            raise ValueError('lightcurve must be a LightCurve object')
        self._shared_with = {}
        self._name = name
        self._observer = observer
        self._lightcurve = lightcurve
        self._isable = {}
        self._method = 'geocenter'

    @property
    def name(self):
        return self._name

    @property
    def observer(self):
        return self._observer

    @property
    def lightcurve(self):
        return self._lightcurve

    def status(self):
        """Returns if the chord is positive or negative
        """
        im = getattr(self.lightcurve, 'immersion', None)
        em = getattr(self.lightcurve, 'emersion', None)
        if im is None and em is None:
            return 'negative'
        else:
            return 'positive'

    @property
    def is_able(self):
        """Return a dictionary with the references enabled and disabled for the fit.
        """
        im = getattr(self.lightcurve, 'immersion', None)
        if im is None and 'immersion' in self._isable:
            del(self._isable['immersion'])
        if im is not None and 'immersion' not in self._isable:
            self._isable['immersion'] = True
        em = getattr(self.lightcurve, 'emersion', None)
        if em is None and 'emersion' in self._isable:
            del(self._isable['emersion'])
        if em is not None and 'emersion' not in self._isable:
            self._isable['emersion'] = True
        return self._isable

    def enable(self, *, time=None):
        """Enables a contact point of the curve to be used in the fit.

        Parameters
        ----------
        time : `None`, `str`
            If ``None``, it will enable all contact points.
            If ``'immersion'`` or ``'emersion'``, it will enable respective
            contact point.
        """
        if time not in [None, 'immersion', 'emersion']:
            raise ValueError('time can only be "immersion", "emersion" or None for both values')
        isable = self.is_able
        for key in isable:
            if time is None or time == key:
                isable[key] = True

    def disable(self, *, time=None):
        """Disables a contact point of the curve to be used in the fit.

        Parameters
        ----------
        time : `None`, `str`
            If None, it will disable all contact points.
            If ``'immersion'`` or ``'emersion'``, it will disable respective
            contact point.
        """
        if time not in [None, 'immersion', 'emersion']:
            raise ValueError('time can only be "immersion", "emersion" or None for both values')
        isable = self.is_able
        for key in isable:
            if time is None or time == key:
                isable[key] = False

    def get_fg(self, *, time=None, vel=False):
        """Returns the on-sky position for this chord at a given time.

        This Chord object must be associated to an Occultation to work, since
        it needs the position of the star and an ephemeris.

        Parameters
        ----------
        time : `astropy.time.Time`, `str`
            It must a time or a list of time to calculate the on-sky position.
            If time is ``'immersion'``, ``'emersion'``, ``'start'`` or ``'end'``,
            it will get the respective on-sky position. If ``time=None``, it
            will return the on-sky position for the immersion and emersion times
            if positive or raise an error if negative.

        vel : `bool`
            If True, it will return the on-sky velocity as well.

        Return
        ------
        f, g, [vf, vg] : `float`, `float`, [`float`, `float`]
            The geocentric on-sky Orthographic projection of the object with `f`
            pointing to the celestial North and `g` pointing to the celestial
            East. The respective velocities (vf, vg) are returned if ``vel=True``.
        """
        from sora.observer import Spacecraft

        if 'chordlist' not in self._shared_with:
            raise ValueError('{} must be associated to an Occultation to use this function'.format(self.__class__.__name__))
        occtime = self._shared_with['chordlist']['time']
        star = self._shared_with['chordlist']['star']
        coord = star.get_position(time=occtime, observer='geocenter')
        ephem = self._shared_with['chordlist']['ephem']

        ref_times = {'immersion': 'immersion', 'emersion': 'emersion', 'start': 'initial_time', 'end': 'end_time'}

        if isinstance(time, str) and time in ref_times.keys():
            time = getattr(self.lightcurve, ref_times[time])
        elif time is None:
            if self.status() == 'negative':
                raise ValueError("{} {}'s LightCurve does not have immersion or emersion time".format(
                    self.__class__.__name__, self.name))
            else:
                immersion = self.lightcurve.immersion
                emersion = self.lightcurve.emersion
                time = Time([immersion, emersion])
        else:
            time = Time(time)

        tca_diff = np.array(np.absolute((time - occtime).jd), ndmin=1)
        if any(tca_diff > 0.5):
            warnings.warn('The difference between a given time and the closest approach instant is {:.2f} days. '
                          'This position could not have a physical meaning.'.format(tca_diff.max()))

        if self._method == 'geocenter' and not isinstance(self.observer, Spacecraft):
            ksio1, etao1 = self.observer.get_ksi_eta(time=time, star=coord)
            ksie1, etae1 = ephem.get_ksi_eta(time=time, star=coord)
            f = ksio1 - ksie1
            g = etao1 - etae1
        else:
            pos_star = star.get_position(time=time, observer=self.observer)
            pos_ephem = ephem.get_position(time=time, observer=self.observer)
            _, f, g, = - pos_ephem.transform_to(SkyOffsetFrame(origin=pos_star)).cartesian.xyz.to(u.km).value

        if not vel:
            return f, g

        if self._method == 'geocenter' and not isinstance(self.observer, Spacecraft):
            ksio2, etao2 = self.observer.get_ksi_eta(time=time + 0.1 * u.s, star=coord)
            ksie2, etae2 = ephem.get_ksi_eta(time=time + 0.1 * u.s, star=coord)
            nf = ksio2 - ksie2
            ng = etao2 - etae2
        else:
            pos_star = star.get_position(time=time + 0.1 * u.s, observer=self.observer)
            pos_ephem = ephem.get_position(time=time + 0.1 * u.s, observer=self.observer)
            _, nf, ng, = - pos_ephem.transform_to(SkyOffsetFrame(origin=pos_star)).cartesian.xyz.to(u.km).value

        return f, g, (nf-f)/0.1, (ng-g)/0.1

    def path(self, *, segment='standard', step=1):
        """Returns the on-sky path of this chord.

        This Chord object must be associated to an Occultation to work, since it
        needs the position of the star and an ephemeris.

        Parameters
        ----------
        segment : `str`
            The segment to get the path of the chord. The available options are:

            ``'positive'``: to get the path between the immersion and emersion
            times if the chord is positive.

            ``'negative'``: to get the path between the start and end of
            observation if the chord is negative.

            ``'standard'``: to get the 'positive' path if the chord is positive
            or 'negative' if the chord is negative.

            ``'full'``: to get  the path between the start and end of observation
            independent if the chord is positive or negative.

            ``'outer'``: to get the path outside the 'positive' path, for
            instance between the start and immersion times and between the
            emersion and end times.

            ``'error'``: to get the path corresponding to the error bars. Be aware
            that some of these segments may return more than 1 path, for example
            ``segment='error'``.

        step : `int`, `float`, `str`
            If a number, it corresponds to the step, in seconds, for each point
            of the path.

            This correspond to an approximate value if it is not a multiple of
            the interval of the segment.

            The step can also be equal to ``'exposure'``. In this case, the path
            will return a set of pairs where each pair will have the position
            at the beginning of the exposure and the end of the exposure.


        Return
        ------
        path f, path g : `array`, `array`
            It will return the path separated by `f` and `g`. If ``step='exposure'``,
            it will return f and g intercalated for each exposure.
        """
        if 'chordlist' not in self._shared_with:
            raise ValueError('{} must be associated to an Occultation to use this function'.format(self.__class__.__name__))

        intervals = []
        try:
            immersion = self.lightcurve.immersion
            emersion = self.lightcurve.emersion
            if segment in ['positive', 'standard']:
                intervals.append((immersion, emersion))
            elif segment == 'outer':
                intervals.append((self.lightcurve.initial_time, immersion))
                intervals.append((emersion, self.lightcurve.end_time))
            elif segment == 'error':
                immersion_err = self.lightcurve.immersion_err*u.s
                emersion_err = self.lightcurve.emersion_err*u.s
                intervals.append((immersion-immersion_err, immersion+immersion_err))
                intervals.append((emersion-emersion_err, emersion+emersion_err))
            elif segment == 'full':
                intervals.append((self.lightcurve.initial_time, self.lightcurve.end_time))
            else:
                raise ValueError('segment {} can not be determined for chord {}'.format(segment, self.name))
        except AttributeError:
            if segment in ['negative', 'standard', 'full']:
                intervals.append((self.lightcurve.initial_time, self.lightcurve.end_time))
            else:
                raise ValueError('segment {} can not be determined for chord {}'.format(segment, self.name))
        exposure = False
        if step == 'exposure':
            exposure = True
            if segment == 'error':
                warnings.warn('"exposure" not available for segment "error".')
                exposure = False
                step = 1
            else:
                try:
                    exptime = self.lightcurve.exptime
                except AttributeError:
                    warnings.warn('{} {} does not have "exptime". Setting continuous path'.format(
                        self.__class__.__name__, self.name))
                    exposure = False
                    step = 1
        elif not isinstance(step, (int, float)):
            raise ValueError("step must be a number or the string 'exposure'")

        vals = []
        for interval in intervals:
            if exposure:
                tref = self.lightcurve.tref
                times = tref + self.lightcurve.time*u.s
                times = times[np.where((times >= interval[0]) & (times <= interval[1]))]
                time_beg = times - exptime/2.0*u.s
                time_end = times + exptime/2.0*u.s
                f1, g1 = self.get_fg(time=time_beg)
                f2, g2 = self.get_fg(time=time_end)
                v = np.array([f1, f2, g1, g2])
                nv = v.T.reshape(2*len(f1), 2)
                for i in nv:
                    vals.append(i)
            else:
                dt = interval[1] - interval[0]
                n = int(dt.sec/step)+1
                if n < 2:
                    n = 2
                time = interval[0] + np.linspace(0, dt.sec, n)*u.s
                f, g = self.get_fg(time=time)
                vals.append(f)
                vals.append(g)

        return vals

    def plot_chord(self, *, segment='standard', only_able=False, ax=None, linestyle='-', **kwargs):
        """Plots the on-sky path of this chord.

        This Chord object must be associated to an Occultation to work, since
        it needs the position of the star and an ephemeris.

        Parameters
        ----------
        segment : `str`
            The segment to plot the chord. The available options are:

            ``'positive'``: to get the path between the immersion and emersion
            times if the chord is positive.

            ``'negative'``: to get the path between the start and end of
            observation if the chord is negative.

            ``'standard'``: to get the 'positive' path if the chord is positive
                or 'negative' if the chord is negative.

            ``'full'``: to get  the path between the start and end of observation
            independent if the chord is positive or negative.

            ``'outer'``: to get the path outside the 'positive' path, for
            instance between the start and immersion times and between the
            emersion and end times.

            ``'error'``: to get the path corresponding to the error bars.

        only_able : `bool`
            Plot only the contact points that are able to be used in the fit.
            If ``segment='error'`` it will show only the contact points able. If
            segment is any other, the path will be plotted only if both
            immersion and emersion are able, or it is a negative chord.

        ax : `matplotlib.pyplot.Axes`
            The axes where to make the plot. If None, it will use the default axes.

        linestyle : `str`
            Default linestyle used in `matplotlib.pyplot.plot`.
            The difference is that now it accepts ``linestyle='exposure'``, where
            the plot will be a dashed line corresponding to each exposure.
            The blank space between the lines can be interpreted as 'dead time'.

        **kwargs
            Any other kwarg will be parsed directly by `maplotlip.pyplot.plot`.
            The only difference is that the default linewidth ``lw=2``.

        """
        import matplotlib.pyplot as plt

        ax = ax or plt.gca()
        ax.set_xlabel('f (km)')
        ax.set_ylabel('g (km)')
        ax.axis('equal')
        if 'chordlist' not in self._shared_with:
            raise ValueError('{} must be associated to an Occultation to use this function'.format(self.__class__.__name__))

        exposure = False
        if linestyle == 'exposure':
            exposure = True
            kwargs['linestyle'] = '-'
        else:
            kwargs['linestyle'] = linestyle

        kwargs['lw'] = kwargs.get('lw', 2)

        steps = {True: 'exposure', False: 1}
        vals = self.path(segment=segment, step=steps[exposure])
        label = kwargs.pop('label', None)
        var = []
        if only_able:
            if segment == 'error':
                if self.is_able['immersion']:
                    immersion = self.lightcurve.immersion
                    immersion_err = self.lightcurve.immersion_err * u.s
                    var += ax.plot(*self.get_fg(time=[immersion - immersion_err, immersion + immersion_err]), **kwargs)
                if self.is_able['emersion']:
                    emersion = self.lightcurve.emersion
                    emersion_err = self.lightcurve.emersion_err * u.s
                    var += ax.plot(*self.get_fg(time=[emersion - emersion_err, emersion + emersion_err]), **kwargs)
            else:
                if self.is_able.get('immersion') in [None, True] and self.is_able.get('emersion') in [None, True]:
                    var += ax.plot(*vals, **kwargs)
        else:
            var += ax.plot(*vals, **kwargs)
        if len(var) > 0:
            var[0].set_label(label)

    def get_impact_param(self, center_f=0, center_g=0, verbose=True):
        """Gets the impact parameter, minimal distance between the chord and the
        centre position.

        This Chord object must be associated to an Occultation to work, since it
        needs the position of the star and an ephemeris.

        Parameters
        ----------
        center_f : `int`, `float`, default=0
            The coordinate in f of the ellipse center;

        center_g : `int`, `float`, default=0
            The coordinate in g of the ellipse center.

        verbose : `bool`, default=True
            If True, prints the obtained values.


        Returns
        -------
        impact, sense : `list`
            The impact parameter (in km) and the direction of the chord relative
            the ellipse center, North (N), South (S), East (E) and West (W).
        """
        f, g = self.path(segment='full')
        r = np.sqrt((f - center_f)**2 + (g - center_g)**2)
        impact = r.min()
        sense = 'NE'
        if g[np.argmin(r)] < center_g:
            sense = sense.replace('N', 'S')
        if f[np.argmin(r)] < center_f:
            sense = sense.replace('E', 'W')
        if verbose:
            print(self.name)
            print('Impact parameter', np.round(impact, 1), sense)
        return impact, sense

    def get_theoretical_times(self, equatorial_radius, center_f=0, center_g=0, oblateness=0, position_angle=0, sigma=0,
                              step=1, verbose=True):
        """Gets the theoretical times and chord size for a given ellipse.

        This Chord object must be associated to an Occultation to work, since
        it needs the position of the star and an ephemeris.

        Parameters
        ----------
        equatorial_radius : `int`, `float`
            The Equatorial radius (semi-major axis) of the ellipse.

        center_f : `int`, `float`, default=0
            The coordinate in f of the ellipse center

        center_g : `int`, `float`, default=0
            The coordinate in g of the ellipse center

        oblateness : `int`, `float`, default=0
            The oblateness of the ellipse.

        position_angle : `int`, `float`, default=0
            The pole position angle of the ellipse in degrees.
            Zero is in the North direction ('g-positive'). Positive clockwise.

        sigma : `int`, `float`
            Uncertainty of the expected ellipse, in km.

        step : `int`, `float`
            Time resolution of the chord, in seconds.

        verbose : `bool`, default=True
            If True, prints the obtained values.


        Returns
        -------
        theory_immersion_time, theory_emersion_time, theory_chord_size : `list`
            The expected immersion time for the given ellipse, the expected
            emersion time for the given ellipse, and the expected chord size
            for the given ellipse.
        """
        from sora.extra import get_ellipse_points

        time_all = Time(np.arange(self.lightcurve.initial_time.jd, self.lightcurve.end_time.jd, step*u.s.to('d')), format='jd')

        f_all, g_all = self.get_fg(time=time_all)
        df_all = (f_all - center_f)
        dg_all = (g_all - center_g)

        r_all = np.sqrt(df_all**2 + dg_all**2)
        cut = r_all < 1.5*equatorial_radius
        df_path = df_all[cut]
        dg_path = dg_all[cut]
        r_path = r_all[cut]
        time = time_all[cut]

        theta_path = np.arctan2(dg_path, df_path)

        f_ellipse, g_ellipse, r_ellipse, theta = get_ellipse_points(theta_path,
                                                                    equatorial_radius=equatorial_radius,
                                                                    center_f=center_f,
                                                                    center_g=center_g,
                                                                    oblateness=oblateness,
                                                                    position_angle=position_angle)

        ev = r_path < r_ellipse + sigma
        if not np.any(ev):
            if verbose:
                print(self.name)
                print('Negative chord \n')
        try:
            imm = time[ev].jd.argmin()
            eme = time[ev].jd.argmax()
            df_chord = [df_path[ev].min(), df_path[ev].max()]
            dg_chord = [dg_path[ev].min(), dg_path[ev].max()]
            theory_chord_size = np.sqrt((df_chord[1]-df_chord[0])**2 + (dg_chord[1]-dg_chord[0])**2)
            theory_immersion_time = time[ev][imm]
            theory_emersion_time = time[ev][eme]
            if verbose:
                print(self.name)
                print('IMMERSION TIME: {} UTC'.format(theory_immersion_time.iso))
                print('EMERSION  TIME: {} UTC'.format(theory_emersion_time.iso))
                print('CHORD SIZE    : {:.3f} km \n'.format(theory_chord_size))
        except:
            theory_chord_size = 0
            theory_immersion_time = None
            theory_emersion_time = None
        return theory_immersion_time, theory_emersion_time, theory_chord_size

    def __repr__(self):
        """String representation of the Chord Class
        """
        return '<{}: {}>'.format(self.__class__.__name__, self.name)

    def __str__(self):
        """String of the Chord Class used in str(obj) or print(obj)
        """
        from sora.observer import Observer

        string = ['-' * 79, self.observer.__str__()]
        if 'chordlist' in self._shared_with and isinstance(self.observer, Observer):
            star = self._shared_with['chordlist']['star']
            occtime = self._shared_with['chordlist']['time']
            coord = star.get_position(time=occtime, observer=self.observer)
            ephem_altaz = self.observer.altaz(self.lightcurve.time_mean, coord)
            string.append('Target altitude: {:.1f} deg'.format(ephem_altaz[0]))
            string.append('Target azimuth:  {:.1f} deg'.format(ephem_altaz[1]))
        string.append('')
        string.append(self.lightcurve.__str__())
        return '\n'.join(string)



import warnings

import astropy.units as u
import numpy as np
from astropy.coordinates import SkyCoord, Longitude, Latitude
from astropy.time import Time

from sora.config import input_tests
from .meta import BaseBody, PhysicalData
from .utils import search_sbdb, search_satdb, apparent_magnitude

__all__ = ['Body']


class Body(BaseBody):
    """Class that contains and manages the information of the body.

    Attributes
    ----------
    name : `str`, required
        The name of the object. It can be the used `spkid` or `designation
        number` to query the SBDB (Small-Body DataBase). In this case, the name
        is case insensitive.

    database : `str`, optional, default='auto'
        The database to query the object. It can be ``satdb`` for our temporary
        hardcoded satellite database, or ``'sbdb'`` to query on the SBDB. If
        database is set as ``auto`` it will try first with ``satdb``,
        then ``sbdb``. If the user wants to use their own information,
        database must be given as ``None``. In this case, `spkid` parameter
        must be given.

    ephem : `sora.EphemKernel`, `sora.EphemHorizons`, `sora.EphemJPL`, `sora.EphemPlanete`
        An Ephem Class that contains information about the ephemeris. It can be
        "horizons" to automatically defined an EphemHorizons object or a list of
        kernels to automatically define an EphemKernel object.

    orbit_class : `str`
        It defines the Orbital class of the body. It can be ``TNO``,
        ``Satellite``, ``Centaur``, ``comet``, ``asteroid``, ``trojan``, ``neo``,
        and ``planet``. It is important for a better characterization of the
        object. If a different value is given, it will be defined as
        ``unclassified``.

    spkid : `str`, `int`, `float`
        If ``database=None``, the user must give a `spkid` or an `ephem`
        which has the `spkid` parameter.

    albedo : `float`, `int`
        The albedo of the object.

    H : `float`, `int`
        The absolute magnitude.

    G : `float`, `int`
        The phase slope.

    diameter : `float`, `int`, `astropy.quantity.Quantity`
        The diameter of the object, in km.

    density : `float`, `int`, `astropy.quantity.Quantity`
        The density of the object, in g/cm³.

    GM : `float`, `int`, `astropy.quantity.Quantity`
        The Standard Gravitational Parameter, in km³/s².

    rotation : `float`, `int`, `astropy.quantity.Quantity`
        The Rotation of the object, in hours.

    pole : `str`, `astropy.coordinates.SkyCoord`
        The Pole coordinates of the object. It can be a `SkyCoord object` or a
        string in the format ``'hh mm ss.ss +dd mm ss.ss'``.

    BV : `float`, `int`
        The B-V color.

    UB : `float`, `int`
        The U-B color.

    smass : `str`
        The spectral type in SMASS classification.

    tholen : `str`
        The spectral type in Tholen classification.

    Note
    ----
    The following attributes are are returned from the Small-Body DataBase when
    ``database='sbdb'`` or from our temporary hardcoded Satellite DataBase when
    ``database='satdb'``:

    `orbit_class`, `spkid`, `albedo`, `H`, `G`, `diameter`, `density`, `GM`,
    `rotation`, `pole`, `BV`, `UB`, `smass`, and `tholen`.

    These are physical parameters the user can give to the object. If a query is
    made and user gives a parameter, the parameter given by the user is defined
    in the *Body* object.

    """

    def __init__(self, name, database='auto', **kwargs):

        allowed_kwargs = ["albedo", "H", "G", "diameter", "density", "GM", "rotation", "pole", "BV", "UB", "smass",
                          "orbit_class", "spkid", "tholen", "ephem"]
        input_tests.check_kwargs(kwargs, allowed_kwargs=allowed_kwargs)
        self._shared_with = {'ephem': {}, 'occultation': {}}
        if database not in ['auto', 'satdb', 'sbdb', None]:
            raise ValueError(f'{database} is not a valid database argument.')
        if database is None:
            self.__from_local(name=name, spkid=kwargs.get('spkid'))
        if database in ['auto', 'satdb']:
            try:
                self.__from_satdb(name=name)
            except ValueError:
                pass
            else:
                database = 'satdb'
        if database in ['auto', 'sbdb']:
            try:
                self.__from_sbdb(name=name)
            except ValueError:
                pass
            else:
                database = 'sbdb'
        if database == 'auto':
            raise ValueError('Object was not located on satdb or sbdb.')
        # set the physical parameters based on the kwarg name.
        if 'smass' in kwargs:
            self.spectral_type['SMASS']['value'] = kwargs.pop('smass')
        if 'tholen' in kwargs:
            self.spectral_type['Tholen']['value'] = kwargs.pop('tholen')
        for key in kwargs:
            setattr(self, key, kwargs[key])
        self._shared_with['ephem']['search_name'] = self._search_name
        self._shared_with['ephem']['id_type'] = self._id_type

    def __from_sbdb(self, name):
        """Searches the object in the SBDB and defines its physical parameters.

        Parameters
        ----------
        name : `str`
            The `name`, `spkid` or `designation number` of the Small Body.

        """
        sbdb = search_sbdb(name)
        self.meta_sbdb = sbdb
        self.name = sbdb['object']['fullname']
        self.shortname = sbdb['object'].get('shortname', self.name)
        self.orbit_class = sbdb['object']['orbit_class']['name']

        pp = sbdb['phys_par']  # get the physical parameters (pp) of the sbdb

        self.albedo = PhysicalData('Albedo', pp.get('albedo'), pp.get('albedo_sig'), pp.get('albedo_ref'), pp.get('albedo_note'))
        self.H = PhysicalData('Absolute Magnitude', pp.get('H'), pp.get('H_sig'), pp.get('H_ref'), pp.get('H_note'), unit=u.mag)
        self.G = PhysicalData('Phase Slope', pp.get('G'), pp.get('G_sig'), pp.get('G_ref'), pp.get('G_note'))
        self.diameter = PhysicalData('Diameter', pp.get('diameter'), pp.get('diameter_sig'), pp.get('diameter_ref'),
                                     pp.get('diameter_note'), unit=u.km)
        self.density = PhysicalData('Density', pp.get('density'), pp.get('density_sig'), pp.get('density_ref'),
                                    pp.get('density_note'), unit=u.g/u.cm**3)
        self.GM = PhysicalData('Standard Gravitational Parameter', pp.get('GM'), pp.get('GM_sig'), pp.get('GM_ref'),
                               pp.get('GM_note'), unit=u.km**3/u.s**2)
        self.rotation = PhysicalData('Rotation', pp.get('rot_per'), pp.get('rot_per_sig'), pp.get('rot_per_ref'),
                                     pp.get('rot_per_note'), unit=u.h)
        self.BV = PhysicalData('B-V color', pp.get('BV'), pp.get('BV_sig'), pp.get('BV_ref'), pp.get('BV_note'))
        self.UB = PhysicalData('U-B color', pp.get('UB'), pp.get('UB_sig'), pp.get('UB_ref'), pp.get('UB_note'))
        if 'pole' in pp:
            self.pole = SkyCoord(pp['pole'].replace('/', ' '), unit=('deg', 'deg'))
            self.pole.ra.uncertainty = Longitude(pp['pole_sig'].split('/')[0], unit=u.deg)
            self.pole.dec.uncertainty = Latitude(pp['pole_sig'].split('/')[1], unit=u.deg)
            self.pole.reference = pp['pole_ref'] or ""
            self.pole.notes = pp['pole_note'] or ""
        else:
            self.pole = None
        self.spectral_type = {
            "SMASS": {"value": pp.get('spec_B'), "reference": pp.get('spec_B_ref'), "notes": pp.get('spec_B_note')},
            "Tholen": {"value": pp.get('spec_T'), "reference": pp.get('spec_T_ref'), "notes": pp.get('spec_T_note')}}
        self.spkid = sbdb['object']['spkid']
        self._des_name = sbdb['object']['des']
        self.discovery = "Discovered {} by {} at {}".format(sbdb['discovery'].get('date'), sbdb['discovery'].get('who'),
                                                            sbdb['discovery'].get('location'))

    def __from_satdb(self, name):
        satdb = search_satdb(name)
        self.name = name.capitalize()
        self.shortname = name.capitalize()
        self.orbit_class = 'satellite'

        self.albedo = PhysicalData('Albedo', *satdb.get('albedo', [None, None, None]))
        self.H = PhysicalData('Absolute Magnitude', *satdb.get('H', [None, None, None]), unit=u.mag)
        self.G = PhysicalData('Phase Slope', *satdb.get('G', [None, None, None]))
        self.diameter = PhysicalData('Diameter', *satdb.get('diameter', [None, None, None]), unit=u.km)
        self.density = PhysicalData('Density', *satdb.get('density', [None, None, None]), unit=u.g / u.cm ** 3)
        self.GM = PhysicalData('Standard Gravitational Parameter', *satdb.get('GM', [None, None, None]),
                               unit=u.km ** 3 / u.s ** 2)
        self.rotation = PhysicalData('Rotation', *satdb.get('rotation', [None, None, None]), unit=u.h)
        if 'pole' in satdb:
            self.pole = SkyCoord(satdb['pole'][0].replace('/', ' '), unit=('deg', 'deg'))
            self.pole.ra.uncertainty = Longitude(satdb['pole'][1].split('/')[0], unit=u.deg)
            self.pole.dec.uncertainty = Latitude(satdb['pole'][1].split('/')[1], unit=u.deg)
            self.pole.reference = satdb['pole'][2] or ""
            self.pole.notes = ""
        else:
            self.pole = None
        self.BV = None
        self.UB = None
        self.spectral_type = {
            "SMASS": {"value": None, "reference": "", "notes": ""},
            "Tholen": {"value": None, "reference": "", "notes": ""}}
        self.spkid = satdb['spkid']
        self._des_name = name
        self.discovery = ""

    def __from_local(self, name, spkid):
        """Defines Body object with default values for mode='local'.
        """
        self.name = name
        self.shortname = name
        self.orbit_class = None
        if not spkid:
            raise ValueError("'spkid' must be given.")
        self.spkid = spkid
        self.albedo = None
        self.H = None
        self.G = None
        self.diameter = None
        self.density = None
        self.GM = None
        self.rotation = None
        self.pole = None
        self.BV = None
        self.UB = None
        self.spectral_type = {"SMASS": {"value": None, "reference": None, "notes": None},
                              "Tholen": {"value": None, "reference": None, "notes": None}}
        self.discovery = ""

    def get_pole_position_angle(self, time, observer='geocenter'):
        """Returns the pole position angle and the aperture angle relative to
        the geocenter.

        Parameters
        ----------
        time : `str`, `astropy.time.Time`
            Time from which to calculate the position.
            It can be a string in the ISO format (yyyy-mm-dd hh:mm:ss.s) or an astropy Time object.

        observer : `str`, `sora.Observer`, `sora.Spacecraft`
            IAU code of the observer (must be present in given list of kernels),
            a SORA observer object or a string: ['geocenter', 'barycenter']

        Returns
        -------
        position_angle, aperture_angle : `float` array
            Position angle and aperture angle of the object's pole, in degrees.
        """
        time = Time(time)
        pole = self.pole
        if np.isnan(pole.ra):
            raise ValueError("Pole coordinates are not defined")
        obj = self.ephem.get_position(time, observer=observer)
        position_angle = obj.position_angle(pole).rad*u.rad
        aperture_angle = np.arcsin(
            -(np.sin(pole.dec)*np.sin(obj.dec) +
              np.cos(pole.dec)*np.cos(obj.dec)*np.cos(pole.ra-obj.ra))
            )
        return position_angle.to('deg'), aperture_angle.to('deg')

    def apparent_magnitude(self, time, observer='geocenter'):
        """Calculates the object's apparent magnitude.

        Parameters
        ----------
        time :  `str`, `astropy.time.Time`
            Reference time to calculate the object's apparent magnitude.
            It can be a string in the ISO format (yyyy-mm-dd hh:mm:ss.s) or an astropy Time object.

        observer : `str`, `sora.Observer`, `sora.Spacecraft`
            IAU code of the observer (must be present in given list of kernels),
            a SORA observer object or a string: ['geocenter', 'barycenter']

        Returns
        -------
        ap_mag : `float`
            Object apparent magnitude.
        """
        from astroquery.jplhorizons import Horizons

        time = Time(time)

        if np.isnan(self.H) or np.isnan(self.G):
            from sora.observer import Observer, Spacecraft
            warnings.warn('H and/or G is not defined for {}. Searching into JPL Horizons service'.format(self.shortname))
            origins = {'geocenter': '@399', 'barycenter': '@0'}
            location = origins.get(observer)
            if not location and isinstance(observer, str):
                location = observer
            if isinstance(observer, (Observer, Spacecraft)):
                location = f'{getattr(observer, "code", "")}@{getattr(observer, "spkid", "")}'
            if not location:
                raise ValueError("observer must be 'geocenter', 'barycenter' or an observer object.")
            obj = Horizons(id=self._search_name, id_type=self._id_type, location=location, epochs=time.jd)
            eph = obj.ephemerides(extra_precision=True)
            if 'H' in eph.keys():
                self.H = eph['H'][0]
                self.H.reference = "JPL Horizons"
                self.G = eph['G'][0]
                self.G.reference = "JPL Horizons"
            if len(eph['V']) == 1:
                return eph['V'][0]
            else:
                return eph['V'].tolist()

        else:
            obs_obj = self.ephem.get_position(time, observer=observer)
            sun_obj = self.ephem.get_position(time, observer='10')

            # Calculates the phase angle between the 2-vectors
            unit_vector_1 = -obs_obj.cartesian.xyz / np.linalg.norm(obs_obj.cartesian.xyz)
            unit_vector_2 = -sun_obj.cartesian.xyz / np.linalg.norm(sun_obj.cartesian.xyz)
            dot_product = np.dot(unit_vector_1, unit_vector_2)
            phase = np.arccos(dot_product).to(u.deg).value
            return apparent_magnitude(self.H.value, self.G.value, obs_obj.distance.to(u.AU).value,
                                      sun_obj.distance.to(u.AU).value, phase)

    def __str__(self):
        from .values import smass, tholen
        out = ['#' * 79 + '\n{:^79s}\n'.format(self.name) + '#' * 79 + '\n',
               'Object Orbital Class: {}\n'.format(self.orbit_class)]
        if self.spectral_type['Tholen']['value'] or self.spectral_type['SMASS']['value']:
            out += 'Spectral Type:\n'
            value = self.spectral_type['SMASS']['value']
            if value:
                out.append('    SMASS: {}  [Reference: {}]\n'.format(value, self.spectral_type['SMASS']['reference']))
            value = self.spectral_type['Tholen']['value']
            if value:
                out.append('    Tholen: {} [Reference: {}]\n'.format(value, self.spectral_type['Tholen']['reference']))
            out += " "*7 + (smass.get(self.spectral_type['SMASS']['value']) or
                            tholen.get(self.spectral_type['Tholen']['value'])) + "\n"
        out.append(self.discovery)

        out.append('\n\nPhysical parameters:\n')
        out.append(self.diameter.__str__())
        out.append(self.mass.__str__())
        out.append(self.density.__str__())
        out.append(self.rotation.__str__())
        if not np.isnan(self.pole.ra):
            out.append('Pole\n    RA:{} +/- {}\n    DEC:{} +/- {}\n    Reference: {}, {}\n'.format(
                       self.pole.ra.__str__(), self.pole.ra.uncertainty.__str__(), self.pole.dec.__str__(),
                       self.pole.dec.uncertainty.__str__(), self.pole.reference, self.pole.notes))
        out.append(self.H.__str__())
        out.append(self.G.__str__())
        out.append(self.albedo.__str__())
        out.append(self.BV.__str__())
        out.append(self.UB.__str__())
        if hasattr(self, 'ephem'):
            out.append('\n' + self.ephem.__str__() + '\n')
        return ''.join(out)



import astropy.units as u
import numpy as np
from astropy.time import Time

from sora.config.decorators import deprecated_alias

__all__ = ['fit_ellipse']


@deprecated_alias(pos_angle='position_angle', dpos_angle='dposition_angle', log='verbose')  # remove this line for v1.0
def fit_ellipse(*args, equatorial_radius, dequatorial_radius=0, center_f=0, dcenter_f=0, center_g=0,
                dcenter_g=0, oblateness=0, doblateness=0, position_angle=0, dposition_angle=0,
                loop=10000000, number_chi=10000, dchi_min=None, verbose=False, ellipse_error=0, sigma_result=1):
    """Fits an ellipse to given occultation using given parameters.

    Parameters
    ----------
    center_f : `int`, `float`, default=0
        The coordinate in f of the ellipse center.

    center_g : `int`, `float`, default=0
        The coordinate in g of the ellipse center.

    equatorial_radius : `int`, `float`
        The Equatorial radius (semi-major axis) of the ellipse.

    oblateness : `int`, `float`, default=0
        The oblateness of the ellipse.

    position_angle : `int`, `float`, default=0
        The pole position angle of the ellipse in degrees.
        Zero is in the North direction ('g-positive'). Positive clockwise.

    dcenter_f : `int`, `float`
        Interval for coordinate f of the ellipse center.

    dcenter_g : `int`, `float`
        Interval for coordinate g of the ellipse center.

    dequatorial_radius `int`, `float`
        Interval for the Equatorial radius (semi-major axis) of the ellipse.

    doblateness : `int`, `float`
        Interval for the oblateness of the ellipse

    dposition_angle : `int`, `float`
        Interval for the pole position angle of the ellipse in degrees.

    loop : `int`, default=10000000
        The number of ellipses to attempt fitting.

    dchi_min : `int`, `float`
        If given, it will only save ellipsis which chi square are smaller than
        chi_min + dchi_min.

    number_chi : `int`, default=10000
        If dchi_min is given, the procedure is repeated until number_chi is
        reached.

    verbose : `bool`, default=False
        If True, it prints information while fitting.

    ellipse_error : `int`, `float`
        Model uncertainty to be considered in the fit, in km.

    sigma_result : `int`, `float`
        Sigma value to be considered as result.

    Returns
    -------
    chisquare : `sora.ChiSquare`
        A ChiSquare object with all parameters.

    Important
    ---------
    Each occultation is added as the first argument(s) directly.

    Mandatory input parameters: 'center_f', 'center_g', 'equatorial_radius',
    'oblateness', and 'position_angle'.

    Parameters fitting interval: 'dcenter_f', 'dcenter_g', 'dequatorial_radius',
    'doblateness', and 'dposition_angle'. Default values are set to zero.
    Search done between (value - dvalue) and (value + dvalue).


    Examples
    --------
    To fit the ellipse to the chords of occ1 Occultation object:

    >>> fit_ellipse(occ1, **kwargs)

    To fit the ellipse to the chords of occ1 and occ2 Occultation objects together:

    >>> fit_ellipse(occ1, occ2, **kwargs)
    """
    from sora.extra import ChiSquare
    from sora.config.visuals import progressbar_show
    from astropy.coordinates import Angle
    from .core import Occultation

    v = {'dcenter_f': dcenter_f, 'dcenter_g': dcenter_g, 'doblateness': doblateness, 'dposition_angle': dposition_angle,
         'dequatorial_radius': dequatorial_radius, 'ellipse_error': ellipse_error, 'sigma_result': sigma_result,
         'dchi_min': dchi_min}
    for key, item in v.items():
        if item is not None and item < 0:
            raise ValueError("{} must be a positive number.".format(key))

    values = []
    chord_name = []
    if len(args) == 0:
        raise ValueError('No occultation have been given as input.')
    for occ in args:
        if not isinstance(occ, Occultation):
            raise TypeError('Given argument must be an Occultation object.')
        for name, chord in occ.chords.items():
            if chord.status() == 'positive':
                if chord.is_able['immersion']:
                    f, g, vf, vg = chord.get_fg(time='immersion', vel=True)
                    err = np.linalg.norm([vf, vg])*chord.lightcurve.immersion_err
                    values.append([f, g, err])
                    chord_name.append(name + '_immersion')
                if chord.is_able['emersion']:
                    f, g, vf, vg = chord.get_fg(time='emersion', vel=True)
                    err = np.linalg.norm([vf, vg])*chord.lightcurve.emersion_err
                    values.append([f, g, err])
                    chord_name.append(name + '_emersion')

    controle_f0 = Time.now()
    f0_chi = np.array([])
    g0_chi = np.array([])
    a_chi = np.array([])
    obla_chi = np.array([])
    posang_chi = np.array([])
    chi2_best = np.array([])

    while len(f0_chi) < number_chi:
        progressbar_show(len(f0_chi), number_chi, prefix='Ellipse fit:')
        chi2 = np.zeros(loop)
        f0 = center_f + dcenter_f*(2*np.random.random(loop) - 1)
        g0 = center_g + dcenter_g*(2*np.random.random(loop) - 1)
        a = equatorial_radius + dequatorial_radius*(2*np.random.random(loop) - 1)
        obla = oblateness + doblateness*(2*np.random.random(loop) - 1)
        obla[obla < 0], obla[obla > 1] = 0, 1
        phi_deg = position_angle + dposition_angle*(2*np.random.random(loop) - 1)
        controle_f1 = Time.now()

        for fi, gi, si in values:
            b = a - a*obla
            phi = phi_deg*(np.pi/180.0)
            dfi = fi-f0
            dgi = gi-g0
            theta = np.arctan2(dgi, dfi)
            ang = theta+phi
            r_model = (a*b)/np.sqrt((a*np.sin(ang))**2 + (b*np.cos(ang))**2)
            f_model = f0 + r_model*np.cos(theta)
            g_model = g0 + r_model*np.sin(theta)
            chi2 += ((fi - f_model)**2 + (gi - g_model)**2)/(si**2 + ellipse_error**2)

        controle_f2 = Time.now()
        if dchi_min is not None:
            region = np.where(chi2 < chi2.min() + dchi_min)[0]
        else:
            region = np.arange(len(chi2))
        chi2_best = np.append(chi2_best, chi2[region])
        if verbose:
            print('Elapsed time: {:.3f} seconds.'.format((controle_f2 - controle_f1).sec))
            print(len(chi2[region]), len(chi2_best))
        f0_chi = np.append(f0_chi, f0[region])
        g0_chi = np.append(g0_chi, g0[region])
        a_chi = np.append(a_chi, a[region])
        obla_chi = np.append(obla_chi, obla[region])
        posang_chi = np.append(posang_chi, phi_deg[region])

    progressbar_show(number_chi, number_chi, prefix='Ellipse fit:')
    chisquare = ChiSquare(chi2_best, len(values), center_f=f0_chi, center_g=g0_chi, equatorial_radius=a_chi,
                          oblateness=obla_chi, position_angle=posang_chi)
    controle_f4 = Time.now()
    if verbose:
        print('Total elapsed time: {:.3f} seconds.'.format((controle_f4 - controle_f0).sec))

    result_sigma = chisquare.get_nsigma(sigma=sigma_result)
    a = result_sigma['equatorial_radius'][0]
    f0 = result_sigma['center_f'][0]
    g0 = result_sigma['center_g'][0]
    obla = result_sigma['oblateness'][0]
    phi_deg = result_sigma['position_angle'][0]
    radial_dispersion = np.array([])
    error_bar = np.array([])
    position_angle_point = np.array([])

    for fi, gi, si in values:
        b = a - a*obla
        phi = phi_deg*(np.pi/180.0)
        dfi = fi-f0
        dgi = gi-g0
        r = np.sqrt(dfi**2 + dgi**2)
        theta = np.arctan2(dgi, dfi)
        ang = theta+phi
        r_model = (a*b)/np.sqrt((a*np.sin(ang))**2 + (b*np.cos(ang))**2)
        radial_dispersion = np.append(radial_dispersion, r - r_model)
        error_bar = np.append(error_bar, si)
        position_angle_point = np.append(position_angle_point, Angle(90*u.deg - theta*u.rad).wrap_at(360 * u.deg).degree)

    for occ in args:
        if isinstance(occ, Occultation):
            occ.fitted_params = {i: result_sigma[i] for i in ['equatorial_radius', 'center_f', 'center_g',
                                                              'oblateness', 'position_angle']}
            occ.chi2_params = {'chord_name': chord_name, 'radial_dispersion': radial_dispersion,
                               'position_angle': position_angle_point, 'radial_error': error_bar,
                               'chi2_min': chisquare.get_nsigma(sigma=sigma_result)['chi2_min'],
                               'nparam': chisquare.nparam, 'npts': chisquare.npts}
    return chisquare



import astropy.units as u
import numpy as np

__all__ = ['filter_negative_chord', 'positionv']


def positionv(star, ephem, observer, time):
    """Calculates the position and velocity of the occultation shadow relative
    to the observer.

    Parameters
    ----------
    star : `sora.Star`
        The coordinate of the star in the same reference frame as the ephemeris.
        It must be a Star object.

    ephem : `sora.Ephem`
        The object ephemeris. It must be an Ephemeris object.

    observer : `sora.Observer`
        The Observer information. It must be an Observer object.

    time : `astropy.time.Time`
        Reference instant to calculate position and velocity. It can be a string
        in the ISO format (yyyy-mm-dd hh:mm:ss.s) or an astropy Time object.

    Returns
    -------
    f, g, vf, vg : `list`
        The orthographic projection of the shadow relative to the observer.
        ``'f'`` is in the x-axis (East-West direction; East positive).
        ``'g'`` is in the y-axis (North-South direction; North positive).
    """
    from sora.ephem import EphemPlanete, EphemJPL, EphemKernel, EphemHorizons
    from sora.observer import Observer
    from sora.star import Star
    from astropy.time import Time

    if type(star) != Star:
        raise ValueError('star must be a Star object')
    if type(ephem) not in [EphemPlanete, EphemJPL, EphemKernel, EphemHorizons]:
        raise ValueError('ephem must be an Ephemeris object')
    if type(observer) != Observer:
        raise ValueError('observer must be an Observer object')
    time = Time(time)

    coord = star.geocentric(time)
    dt = 0.1*u.s

    if type(ephem) == EphemPlanete:
        ephem.fit_d2_ksi_eta(coord, verbose=False)
    ksio1, etao1 = observer.get_ksi_eta(time=time, star=coord)
    ksie1, etae1 = ephem.get_ksi_eta(time=time, star=coord)

    f = ksio1-ksie1
    g = etao1-etae1

    ksio2, etao2 = observer.get_ksi_eta(time=time+dt, star=coord)
    ksie2, etae2 = ephem.get_ksi_eta(time=time+dt, star=coord)

    nf = ksio2-ksie2
    ng = etao2-etae2

    vf = (nf-f)/0.1
    vg = (ng-g)/0.1

    return f, g, vf, vg


def filter_negative_chord(chord, chisquare, step=1, sigma=0):
    """Get points for the ellipse with the given input parameters.

    Parameters
    ----------
    chord : `sora.observer.Chord`
        Chord object, must be associated to an Occultation to work.

    chisquare : `sora.extra.ChiSquare`
        Resulted ChiSquare object of fit_ellipse.

    sigma : `int`, `float`
        Uncertainty of the expected ellipse, in km.

    step : `int`, `float`, `str`
        If a number, it corresponds to the step, in seconds, for each point of
        the chord path. The step can also be equal to ``'exposure'``. In this
        case, the chord path will consider the lightcurve individual times and
        exptime.
    """
    from sora.config.visuals import progressbar
    from sora.extra import ChiSquare
    from sora.extra.utils import get_ellipse_points

    keep = []
    if step == 'exposure':
        try:
            step = chord.lightcurve.exptime/10.
        except AttributeError:
            raise AttributeError('Chord.lightcurve does not have "exptime"')
        time_all = np.arange(chord.lightcurve.time.min(), chord.lightcurve.time.max(), step)
        time_exposure = np.array([])
        for i in range(len(chord.lightcurve.time)):
            event_model = (time_all > chord.lightcurve.time[i] - chord.lightcurve.exptime/2.) & (
                        time_all < chord.lightcurve.time[i] + chord.lightcurve.exptime/2.)
            time_exposure = np.append(time_exposure, time_all[event_model])
            f_all, g_all = chord.get_fg(time=time_exposure*u.s + chord.lightcurve.tref)
    else:
        f_all, g_all = chord.path(segment='full', step=step)
    for i in progressbar(range(len(chisquare.data['chi2'])), 'Filter chord: {}'.format(chord.name)):
        df_all = (f_all - chisquare.data['center_f'][i])
        dg_all = (g_all - chisquare.data['center_g'][i])

        r_all = np.sqrt(df_all**2 + dg_all**2)
        cut = r_all < 1.5*chisquare.data['equatorial_radius'][i]
        df_path = df_all[cut]
        dg_path = dg_all[cut]
        r_path = r_all[cut]
        theta_path = np.arctan2(dg_path, df_path)

        r_ellipse = get_ellipse_points(theta_path,
                                       equatorial_radius=chisquare.data['equatorial_radius'][i],
                                       oblateness=chisquare.data['oblateness'][i],
                                       center_f=chisquare.data['center_f'][i],
                                       center_g=chisquare.data['center_g'][i],
                                       position_angle=chisquare.data['position_angle'][i])[2]
        keep.append(np.all(r_path - r_ellipse + sigma > 0))

    filtered_chisquare = ChiSquare(chisquare.data['chi2'][keep], chisquare.npts,
                                   center_f=chisquare.data['center_f'][keep],
                                   center_g=chisquare.data['center_g'][keep],
                                   equatorial_radius=chisquare.data['equatorial_radius'][keep],
                                   oblateness=chisquare.data['oblateness'][keep],
                                   position_angle=chisquare.data['position_angle'][keep])
    return filtered_chisquare



from collections import OrderedDict

__all__ = ['List']


class List(OrderedDict):
    """Abstract Class to handle SORA List Classes.

    This Class inherits from OrderedDict, which is a dictionary that keeps the
    order that the objects were added. With this, the user can access a item by
    its order as well.

    """
    _allowed_types = None  # _allowed_types attribute must be defined for each class with the list of allowed types.
    _set_func = "_add_item"  # _set_func attribute must be defined for each class with the name of the function to add item.

    def __setitem__(self, name, item):
        """Overwrites the default __setitem__ from OrderedDict.

        The reason is to forbid the user to define an item as ``obj[name] = item``
        and force the use of the function 'add_item'.
        """
        raise ValueError("{} cannot be set directly. Please use {}".format(self.__class__.__name__, self._set_func))

    # __add_item must be defined inside an add method of Child Class.
    def _add_item(self, name, item):
        """Adds an item to the list with the name given.

        Parameters
        ----------
        name : `str`
            The name of the item. It must be unique for a item in the list.

        item : `_allowed_types`
            The item to be added to the list. Its type must be one of the types
            in the attribute ``self._allowed_type``.
        """
        if not isinstance(name, str):
            raise TypeError("name must be a string")
        if name == '':
            raise ValueError("name can not be an empty string")
        if name in self.keys():
            raise ValueError("{} already exists. It must be removed before redefined".format(name))
        if not isinstance(item, self._allowed_types):
            raise TypeError("item must be an allowed type: {}".format([k.__name__ for k in self._allowed_types]))
        super().__setitem__(name, item)

    def __getitem__(self, key):
        """"Defines behaviour for ``obj[name]``.

        Parameters
        ----------
        key : `str`, `int`
            The key used to find item in the list. It can the name given in
            `_add_item` or the number corresponding to the order the item is
            stored in the list.
        """
        if isinstance(key, int):
            try:
                key = list(self.keys())[key]
            except IndexError:
                raise IndexError("Cannot get item {} for list with size {}".format(key, len(self)))
            return self[key]
        elif isinstance(key, str):
            return super().__getitem__(key)
        else:
            raise TypeError("{} can only be indexed with the "
                            "named keys and integers.".format(self.__class__.__name__))

    def __delitem__(self, key):
        """Defines behaviour of ``del(obj[item])`` to delete only one item in the list.

        Parameters
        ----------
        key : `str`, `int`
            The key used to find item in the list. It can the name given in
            ``_add_item`` or the number corresponding to the order the item is
            stored in the list.
        """
        if isinstance(key, int):
            try:
                key = list(self.keys())[key]
            except IndexError:
                raise IndexError("Cannot get item {} for list with size {}".format(key, len(self)))
        if not isinstance(key, str):
            raise TypeError("{} can only be indexed with the "
                            "named keys and integers.".format(self.__class__.__name__))
        super().__delitem__(key)

    def __str__(self):
        """Defines behaviour for str(obj) or print(obj)
        """
        strings = []
        for key in self.keys():
            strings.append(str(self[key]))
        return "\n".join(strings)

    def __repr__(self):
        """Defines the object's string representation.

        <ClassListName:
            0: Type of item0 (Name of item0)
            1: Type of item1 (Name of item1)
            ...>
        """
        data = ['\n    {}: {}({})'.format(i, self[i].__class__.__name__, self[i].name) for i in range(len(self))]
        return '<{}:{}>'.format(self.__class__.__name__, ''.join(data))



import warnings
import numpy as np

__all__ = ['occ_detect']


def occ_detect(flux, dflux, time, cycle, maximum_duration=None, dur_step=None, snr_limit=None,
               n_detections=None, plot=False):
    """Detects automatically the occultation event in the light curve.

    Detects a 'square well' shaped transit. All parameters are optional.


    Parameters
    ----------
    flux : `float` array
        Flux of the time series. Dependent variable.

    dflux: `float` array
        Error in the flux. Error in the dependent variable.

    time: `float` array
        Time variable. Independent variable.

    cycle: `float`
        Sampling value of the time series.

    maximum_duration : `float`, default: light curve time span
        Maximum duration of the occultation event.

    dur_step : `float`, default: 1/2 cycle
        Step size to sweep occultation duration event.

    snr_limit : `float`, default=None
        Minimum occultation SNR.

    n_detections : `int`, default=1
        Number of detections regardless the SNR. `n_detections` is superseded by
        `snr_limit`.

    plot : `boolean`, default=False
        True if output plots are desired.


    Returns
    -------
    OrderedDict : `dict`
        An ordered dictionary of :attr:`name`::attr:`value` pairs for each
        parameter.


    Examples
    --------
    >>> from sora.lightcurve.occdetect import occ_detect
    >>> params = occ_detect(flux, dflux, time, 0.2)
    >>> params
    {'rank': 1,
    'occultation_duration': 40.1384063065052,
    'central_time': 7916.773870512843,
    'immersion_time': 7896.7046673595905,
    'emersion_time': 7936.843073666096,
    'time_err': 0.05011036992073059,
    'depth': 0.8663887801707082,
    'depth_err': 0.10986223384336465,
    'baseline': 0.9110181732552853,
    'baseline_err': 0.19045768512595365,
    'snr': 7.886138392251848,
    'occ_mask': array([False, False, False, ..., False, False, False])}

    """

    # duration of the light curve
    time_span = time[-1] - time[0]
    if maximum_duration and (maximum_duration > time_span):
        warnings.warn('Occultation duration (maximum_duration={0}) '
                      'exceeds the time series length ({1:0.5f}).'
                      ' maximum_duration reset to the time series length.'
                      .format(maximum_duration, time_span))
        maximum_duration = time_span
    if not maximum_duration:
        maximum_duration = time_span

    if not dur_step:
        dur_step = cycle / 2

    if dur_step < cycle / 2:
        warnings.warn('The given dur_step is oversampled by a factor '
                      'of {0:0.1f} and has been reset to half a cycle '
                      '({1:0.4f}).'.format((cycle / 2.) / dur_step, cycle / 2.))
        dur_step = cycle / 2

    duration_grid = np.arange(dur_step, maximum_duration, dur_step)
    # initial occultation mask (all data points)
    mask = np.ones(len(time), dtype=bool)
    # initial detection rank
    rank = 1

    if snr_limit:
        # minimum SNR accepted in a detection for multiple search
        snr_value = snr_limit + 1
        occ0 = run_bls(flux, dflux, time, time_span, duration_grid)
        mask *= ~occ0['occ_mask']
        while snr_value > snr_limit:
            rank += 1
            occ1 = run_bls(flux, dflux, time, time_span, duration_grid, mask=mask,
                           rank=rank)
            if occ1['snr'] > snr_limit:
                snr_value = occ1['snr']
                mask *= ~occ1['occ_mask']
                occ0 = summarize_bls(occ0, occ1)
            else:
                snr_value = snr_limit

        if plot:
            plot_occ_detect(occ0, flux, time)
        return occ0
    elif n_detections:
        # search the n best fits
        occ0 = run_bls(flux, dflux, time, time_span, duration_grid)
        mask *= ~occ0['occ_mask']
        for i in range(n_detections - 1):
            rank += 1
            occ1 = run_bls(flux, dflux, time, time_span, duration_grid, mask=mask,
                           rank=rank)
            mask *= ~occ1['occ_mask']
            occ0 = summarize_bls(occ0, occ1)

        if plot:
            plot_occ_detect(occ0, flux, time)
        return occ0
    else:
        # search only the first best fit
        occ0 = run_bls(flux, dflux, time, time_span, duration_grid)

        if plot:
            plot_occ_detect(occ0, flux, time)

        return occ0


def plot_occ_detect(occ, flux, time):
    import matplotlib.pyplot as plt

    n = np.size(occ['rank'])
    if n > 1:
        # case for multiple detections
        plt.plot(time, flux, 'k.-')
        mask = np.zeros(len(time), dtype=bool)
        for i in range(n):
            trues = np.sum(occ['occ_mask'][i])
            plt.plot(time[occ['occ_mask'][i]], np.repeat(np.mean(flux[occ['occ_mask'][i]]), trues),
                     '.', label='Rank: ' + str(i + 1))
            mask += occ['occ_mask'][i]

        falses = list(mask).count(False)
        plt.plot(time[~mask], np.repeat(np.mean(flux[~mask]), falses), 'r.', label='Baseline')
        plt.xlabel('Time [seconds]')
        plt.ylabel('Relative Flux')
        plt.legend()

    else:
        # case for single occultation
        trues = list(occ['occ_mask']).count(True)
        falses = list(occ['occ_mask']).count(False)
        plt.plot(time, flux, 'k.-')
        plt.plot(time[occ['occ_mask']], np.repeat(np.mean(flux[occ['occ_mask']]), trues),
                 '.', label='Occultation')
        plt.plot(time[~occ['occ_mask']], np.repeat(np.mean(flux[~occ['occ_mask']]), falses),
                 'r.', label='Baseline')
        plt.xlabel('Time [seconds]')
        plt.ylabel('Relative Flux')
        plt.legend()


def run_bls(flux, dflux, time, per_grid, dur_grid, mask=None, rank=None):
    """ Private function to find the best box fit suitable to the data
    """
    from astropy.timeseries import BoxLeastSquares

    # object with no occultation mask
    mskmodel = BoxLeastSquares(time, flux, dy=dflux)
    # if there is no dflux array, reset it to None in case of
    # using a mask
    if dflux is None:
        dfluxmask = None
    else:
        dfluxmask = dflux[mask]

    # object with occultation mask
    if np.sum(mask):
        model = BoxLeastSquares(time[mask], flux[mask],
                                dy=dfluxmask)
    else:
        model = mskmodel

    r = model.power(per_grid, dur_grid, objective='snr', method='fast')
    # statistics of the BLS fit
    stats = model.compute_stats(r.period, r.duration, r.transit_time)
    # occultation mask of the event with respect to all data
    occ_mask = mskmodel.transit_mask(time, r.period, r.duration,
                                     r.transit_time)
    # parameters computation for clarity purposes
    occultation_duration = r.duration[0]
    central_time = stats['transit_times'][0]
    immersion_time = stats['transit_times'][0] - r.duration[0] / 2
    emersion_time = stats['transit_times'][0] + r.duration[0] / 2
    time_err = np.median(time[1:-1] - time[0:-2]) / 2
    depth = np.mean(flux[~occ_mask]) - np.mean(flux[occ_mask])
    depth_err = np.std(flux[occ_mask], ddof=1)
    baseline = np.mean(flux[~occ_mask])
    baseline_err = np.std(flux[~occ_mask], ddof=1)
    # If there is only one measurement during the occultation it will
    # use the baseline_err to compute SNR, otherwise it will use depth_err
    if np.sum(occ_mask) < 2:
        snr = depth / baseline_err
    else:
        snr = depth / depth_err

    # define rank
    if rank:
        rank = rank
    else:
        rank = 1

    return {'rank': rank,
            'occultation_duration': occultation_duration,
            'central_time': central_time,
            'immersion_time': immersion_time,
            'emersion_time': emersion_time,
            'time_err': time_err,
            'depth': depth, 'depth_err': depth_err,
            'baseline': baseline, 'baseline_err': baseline_err,
            'snr': snr, 'occ_mask': occ_mask}


def summarize_bls(dict1, dict2):
    """ Function to merge dictionaries returned by BLS and
        to keep values of common keys in list.
    """
    dict3 = {}
    for key, value in dict1.items():
        if key == 'occ_mask':
            sz = int(np.size(dict1[key]) / np.size(dict2[key]))
            if sz > 1:
                k = [None] * (sz + 1)
                for i in range(sz):
                    k[i] = dict1[key][i]
                k[i + 1] = dict2[key]
                dict3[key] = k
            else:
                dict3[key] = [dict1[key], dict2[key]]
        else:
            dict3[key] = np.append(dict1[key], dict2[key])
    return dict3



import astropy.units as u
from astropy.coordinates import SkyCoord, EarthLocation, GCRS, AltAz
from astropy.time import Time

from sora.config import input_tests
from .utils import search_code_mpc

__all__ = ['Observer', 'Spacecraft']


class Observer:
    """Defines the observer object.

    Attributes
    ----------
    name : `str`
        Name for the Observer. Observer is uniquely defined (name must be
        different for each observer).

    code : `str`
        The IAU code for SORA to search for its coordinates in MPC database.

    site : `astropy.coordinates.EarthLocation`
        User provides an EarthLocation object.

    lon : `str`, `float`
        The Longitude of the site in degrees.
        Positive to East. Range (0 to 360) or (-180 to +180).
        User can provide in degrees (`float`) or hexadecimal (`string`).

    lat : `str`, `float`
        The Latitude of the site in degrees.
        Positive North. Range (+90 to -90).
        User can provide in degrees (float) or hexadecimal (string).

    height : `int`, `float`
        The height of the site in meters above see level.

    ephem : `str`, `list`
        The ephemeris used to locate the observer on space.
        It can be "horizons" to use horizons or a list of kernels


    Examples
    --------
    User can provide one of the following to define an observer:

    - If user will use the MPC name for the site:

    >>> Observer(code)

    - If user wants to use a different name from the MPC database:

    >>> Observer(name, code)

    - If user wants to use an EarthLocation value:

    >>> from astropy.coordinates import EarthLocation
    >>> EarthLocation(lon, lat, height)
    >>> Observer(name, site)

    - If user wants to give site coordinates directly:

    >>> Observer(name, lon, lat, height)

    """

    def __init__(self, **kwargs):

        input_tests.check_kwargs(kwargs, allowed_kwargs=['code', 'height', 'lat', 'lon', 'name', 'site', 'ephem'])
        self.__name = kwargs.get('name', '')
        if 'code' in kwargs and any(i in kwargs for i in ['lon', 'lat', 'height']):
            raise ValueError("The Observer object is instantiated with IAU code or coordinates, not both.")
        if 'code' in kwargs:
            self.code = kwargs['code']
            try:
                name, self.site = search_code_mpc()[self.code]
                self.__name = kwargs.get('name', name)
            except:
                raise ValueError('code {} could not be located in MPC database'.format(self.code))
        elif 'site' in kwargs:
            self.site = input_tests.test_attr(kwargs['site'], EarthLocation, 'site')
        elif all(i in kwargs for i in ['lon', 'lat']):
            self.site = EarthLocation(kwargs['lon'], kwargs['lat'], kwargs.get('height', 0.0))
        else:
            raise ValueError('Input parameters could not be determined')
        self.ephem = kwargs.get('ephem', 'horizons')

    def get_ksi_eta(self, time, star):
        """Calculates relative position to star in the orthographic projection.

        Parameters
        ----------
        time : `str`, `astropy.time.Time`
            Reference time to calculate the position. It can be a string in the
            format ``'yyyy-mm-dd hh:mm:ss.s'`` or an astropy `Time` object.

        star : `str`, `astropy.coordinates.SkyCoord`
            The coordinate of the star in the same reference frame as the
            ephemeris. It can be a string in the format ``'hh mm ss.s +dd mm ss.ss'``
            or an astropy `SkyCoord` object.


        Returns
        -------
        ksi, eta : `float`
            On-sky orthographic projection of the observer relative to a star.
            ``ksi`` is in the North-South direction (North positive).
            ``eta`` is in the East-West direction (East positive).
        """
        from astropy.coordinates.matrix_utilities import rotation_matrix

        time = input_tests.test_attr(time, Time, 'time')
        try:
            star = SkyCoord(star, unit=(u.hourangle, u.deg))
        except:
            raise ValueError('star is not an astropy object or a string in the format "hh mm ss.s +dd mm ss.ss"')

        itrs = self.site.get_itrs(obstime=time)
        gcrs = itrs.transform_to(GCRS(obstime=time))
        rz = rotation_matrix(star.ra, 'z')
        ry = rotation_matrix(-star.dec, 'y')

        cp = gcrs.cartesian.transform(rz).transform(ry)
        return cp.y.to(u.km).value, cp.z.to(u.km).value

    def sidereal_time(self, time, mode='local'):
        """Calculates the Apparent Sidereal Time at a reference time.

        Parameters
        ----------
        time : `str`, `astropy.time.Time`
            Reference time to calculate sidereal time.It can be a string
            in the ISO format (yyyy-mm-dd hh:mm:ss.s) or an astropy Time object.

        mode : `str`
            Local or greenwich time.
            If mode set ``'local'`` calculates the sidereal time for the
            coordinates of this object.
            If mode set ``'greenwich'`` calculates the Greenwich Apparent
            Sidereal Time.


        Returns
        -------
        sidereal_time
            An Astropy Longitude object with the Sidereal Time.
        """
        # return local or greenwich sidereal time
        time = input_tests.test_attr(time, Time, 'time')
        time.location = self.site
        if mode == 'local':
            return time.sidereal_time('apparent')
        elif mode == 'greenwich':
            return time.sidereal_time('apparent', 'greenwich')
        else:
            raise ValueError('mode must be "local" or "greenwich"')

    def altaz(self, time, coord):
        """Calculates the Altitude and Azimuth at a reference time for a coordinate.

        Parameters
        ----------
        time : `str`, `astropy.time.Time`
            Reference time to calculate the sidereal time. It can be a string
            in the ISO format (yyyy-mm-dd hh:mm:ss.s) or an astropy Time object.

        coord : `str`, `astropy.coordinates.SkyCoord`
            Coordinate of the target ICRS.


        Returns
        -------
        altitude : `float`
            Object altitude in degrees.

        azimuth : `float`
            Object azimuth in degrees.
        """
        time = input_tests.test_attr(time, Time, 'time')
        if type(coord) == str:
            coord = SkyCoord(coord, unit=(u.hourangle, u.deg))
        ephem_altaz = coord.transform_to(AltAz(obstime=time, location=self.site))
        return ephem_altaz.alt.deg, ephem_altaz.az.deg

    def get_vector(self, time, origin='barycenter'):
        """Return the vector Origin -> Observer in the ICRS

        Parameters
        ----------
        time : `str`, `astropy.time.Time`
            Reference time to calculate the object position. It can be a string
            in the ISO format (yyyy-mm-dd hh:mm:ss.s) or an astropy Time object.

        origin : `str`
            Origin of vector. It can be 'barycenter' or 'geocenter'.

        Returns
        -------
        coord : `astropy.coordinates.SkyCoord`
            Astropy SkyCoord object with the vector origin -> observer at the given time.
        """
        from sora.ephem.utils import ephem_horizons, ephem_kernel

        itrs = self.site.get_itrs(obstime=time)
        gcrs = itrs.transform_to(GCRS(obstime=time))

        if self.ephem == 'horizons':
            vector = ephem_horizons(time=time, target=self.spkid, observer=origin, id_type='majorbody', output='vector')
        else:
            vector = ephem_kernel(time=time, target=self.spkid, observer=origin, kernels=self.ephem, output='vector')

        topo = SkyCoord(vector.cartesian + gcrs.cartesian, representation_type='cartesian')
        if not topo.isscalar and len(topo) == 1:
            topo = topo[0]
        return topo

    @property
    def name(self):
        return self.__name

    @property
    def lon(self):
        return self.site.lon

    @lon.setter
    def lon(self, lon):
        lat = self.site.lat
        height = self.site.height
        site = EarthLocation(lon, lat, height)
        self.site = site

    @property
    def lat(self):
        return self.site.lat

    @lat.setter
    def lat(self, lat):
        lon = self.site.lon
        height = self.site.height
        site = EarthLocation(lon, lat, height)
        self.site = site

    @property
    def height(self):
        return self.site.height

    @height.setter
    def height(self, height):
        lon = self.site.lon
        lat = self.site.lat
        site = EarthLocation(lon, lat, height)
        self.site = site

    @property
    def spkid(self):
        return '399'

    def __repr__(self):
        """String representation of the Observer Class
        """
        return '<{}: {}>'.format(self.__class__.__name__, self.name)

    def __str__(self):
        """ String representation of the Observer class
        """
        out = ('Site: {}\n'
               'Geodetic coordinates: Lon: {}, Lat: {}, height: {:.3f}'.format(
                   self.name, self.site.lon.__str__(), self.site.lat.__str__(),
                   self.site.height.to(u.km))
               )
        return out


class Spacecraft:
    """Defines a spacecraft observer object.

    Attributes
    ----------
    name : `str`
        Name for the Observer. Observer is uniquely defined (name must be
        different for each observer).

    spkid : `str`, required
        `spkid` of the targeting object.

    ephem : `str`, `list`
        The ephemeris used to locate the observer on space.
        It can be "horizons" to use horizons or a list of kernels

    """

    def __init__(self, name, spkid, ephem='horizons'):
        self._name = name
        self._spkid = spkid
        self._ephem = ephem

    @property
    def name(self):
        return self._name

    @property
    def spkid(self):
        return self._spkid

    @property
    def ephem(self):
        return self._ephem

    def get_vector(self, time, origin='barycenter'):
        """Return the vector Origin -> Observer in the ICRS

        Parameters
        ----------
        time : `str`, `astropy.time.Time`
            Reference time to calculate the object position. It can be a string
            in the ISO format (yyyy-mm-dd hh:mm:ss.s) or an astropy Time object.

        origin : `str`
            Origin of vector. It can be 'barycenter' or 'geocenter'.

        Returns
        -------
        coord : `astropy.coordinates.SkyCoord`
            Astropy SkyCoord object with the vector origin -> observer at the given time.
        """
        from sora.ephem.utils import ephem_horizons, ephem_kernel
        time = Time(time)

        if self.ephem == 'horizons':
            vector = ephem_horizons(time=time, target=self.spkid, observer=origin, id_type='majorbody', output='vector')
        else:
            vector = ephem_kernel(time=time, target=self.spkid, observer=origin, kernels=self.ephem, output='vector')
        return vector

    def __repr__(self):
        """String representation of the Spacecraft Class
        """
        return '<{}: {}>'.format(self.__class__.__name__, self.name)

    def __str__(self):
        """ String representation of the Spacecraft class
        """
        out = ['Spacecraft: {} (spkid={})'.format(self.name, self.spkid),
               'Positions from {}'.format(str(self.ephem))]
        return '\n'.join(out)



from setuptools import setup, find_packages
from os import path

# read the contents of README.rst
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='sora-astro',
    packages=find_packages(),
    version='0.2rc1',
    license='MIT',
    description='Stellar Occultation Library',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='SORA Team',
    author_email='rio.occteam@gmail.com',
    url='https://github.com/riogroup/SORA',
    keywords=['science', 'astronomy', 'occultation'],
    install_requires=[
        'numpy>=1.18',
        'pyerfa>=2.0',
        'astropy>=4.0',
        'astroquery>=0.4.1',
        'spiceypy>=3.0.2',
        'matplotlib>=3.2',
        'scipy>=1.4.1',
        'requests',
    ],
    python_requires=">=3.7, <4",
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Science/Research ',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.7',
        'Topic :: Scientific/Engineering :: Astronomy',
    ],
    project_urls={
        'Bug Reports': 'https://github.com/riogroup/SORA/issues',
        'Documentation': 'https://sora.readthedocs.io/',
    },
)



import numpy as np

__all__ = ['get_ellipse_points']


def get_ellipse_points(theta, equatorial_radius, oblateness=0.0, center_f=0.0, center_g=0.0,
                       position_angle=0.0):
    """Get points for the ellipse with the given input parameters.

    Parameters
    ----------
    theta : `float` array
        Angular coordinate, in degrees, to return the ellipse points.

    equatorial_radius : `float`, `int`
        Semi-major axis of the ellipse.

    oblateness : `float`, `int`, default=0
        Oblateness of the ellipse.

    center_f : `float`, `int`, default=0
        Coordinate of the ellipse (abscissa).

    center_g : `float`, `int`, default=0
        Coordinate of the ellipse (ordinate).

    position_angle : `float`, `int`, default=0
        The pole position angle of the ellipse in degrees.
        Zero is in the North direction ('g-positive'). Positive clockwise.

    Returns
    -------
    x_model : `float`, array
        Cartesian x-component, in km

    y_model : `float`, array
            Cartesian y-component, in km

    r_model : `float`, array
            Radial distance, in km

    theta : `float` array
            Angular coordinate, in degrees, to return the ellipse points.
    """
    a = equatorial_radius
    b = equatorial_radius - equatorial_radius * oblateness
    phi = position_angle * (np.pi / 180.0)
    ang = theta + phi
    r_model = (a * b) / np.sqrt((a * np.sin(ang)) ** 2 + (b * np.cos(ang)) ** 2)
    x_model = r_model * np.cos(theta) + center_f
    y_model = r_model * np.sin(theta) + center_g
    return x_model, y_model, r_model, theta



import warnings

import astropy.units as u
import numpy as np
from astropy.coordinates import SkyCoord
from astropy.time import Time

from sora.config import input_tests
from sora.config.decorators import deprecated_function


class BaseEphem:
    def __init__(self, name=None, spkid=None, **kwargs):
        # remove 'H', 'G', 'mass' and 'radius' from allowed kwargs and docstring for v1.0
        input_tests.check_kwargs(kwargs, allowed_kwargs=['error_dec', 'error_ra', 'H', 'G', 'mass', 'radius'])
        #
        self._shared_with = {'body': {}, 'occultation': {}}
        self.name = name
        if spkid:
            self.spkid = spkid
            self.code = self.spkid  # remove this line for v1.0
        self.error_ra = u.Quantity(kwargs.get('error_ra', 0), unit=u.arcsec)
        if self.error_ra < 0:
            warnings.warn("Error in RA cannot be negative. Using absolute value.")
            self.error_ra = np.absolute(self.error_ra)
        self.error_dec = u.Quantity(kwargs.get('error_dec', 0), unit=u.arcsec)
        if self.error_dec < 0:
            warnings.warn("Error in DEC cannot be negative. Using absolute value.")
            self.error_dec = np.absolute(self.error_dec)
        self.offset = (0, 0)
        # start of block removal for v1.0
        if 'radius' in kwargs:
            self.radius = kwargs['radius'] * u.km
        self.mass = kwargs.get('mass', 0.0) * u.kg
        if 'H' in kwargs:
            self.H = kwargs['H']
        if 'G' in kwargs:
            self.G = kwargs['G']
        # end of block removal for v1.0

    @property
    def spkid(self):
        if 'spkid' in self._shared_with['body']:
            return self._shared_with['body']['spkid']
        elif hasattr(self, '_spkid'):
            return self._spkid
        else:
            raise AttributeError('{} does not have spkid'.format(self.__class__.__name__))

    @spkid.setter
    def spkid(self, value):
        if 'spkid' in self._shared_with['body']:
            raise AttributeError('When {} is associated to a Body object, spkid must be given to the Body'
                                 ' object.'.format(self.__class__.__name__))
        self._spkid = str(int(value))

    # Start of block removal for v1.0
    @property
    def radius(self):
        if 'radius' in self._shared_with['body']:
            return self._shared_with['body']['radius']
        elif hasattr(self, '_radius'):
            return self._radius
        else:
            raise AttributeError('{} does not have radius'.format(self.__class__.__name__))

    @radius.setter
    def radius(self, value):
        if 'radius' in self._shared_with['body']:
            raise AttributeError('When {} is associated to a Body object, radius must be given to the Body'
                                 ' object.'.format(self.__class__.__name__))
        self._radius = float(value)

    @property
    def H(self):
        if 'H' in self._shared_with['body']:
            return self._shared_with['body']['H'].value
        elif hasattr(self, '_H'):
            return self._H
        else:
            raise AttributeError('{} does not have H'.format(self.__class__.__name__))

    @H.setter
    def H(self, value):
        if 'H' in self._shared_with['body']:
            raise AttributeError('When {} is associated to a Body object, H must be given to the Body'
                                 ' object.'.format(self.__class__.__name__))
        self._H = float(value)

    @property
    def G(self):
        if 'G' in self._shared_with['body']:
            return self._shared_with['body']['G']
        elif hasattr(self, '_G'):
            return self._G
        else:
            raise AttributeError('{} does not have G'.format(self.__class__.__name__))

    @G.setter
    def G(self, value):
        if 'G' in self._shared_with['body']:
            raise AttributeError('When {} is associated to a Body object, G must be given to the Body'
                                 ' object.'.format(self.__class__.__name__))
        self._G = float(value)

    def apparent_magnitude(self, time):
        """Calculates the object's apparent magnitude.

        Parameters
        ----------
        time : `str`, `astropy.time.Time`
            Reference time to calculate the object's apparent magnitude.
            It can be a string in the ISO format (yyyy-mm-dd hh:mm:ss.s) or an astropy Time object.

        Returns
        -------
        ap_mag : `float`
            Object apparent magnitude.
        """
        from sora.body.utils import apparent_magnitude
        from astroquery.jplhorizons import Horizons
        from astropy.coordinates import get_sun

        time = Time(time)

        if getattr(self, 'H', None) is None or getattr(self, 'G', None) is None:
            search_name = self._shared_with['body'].get('search_name', self.name)
            id_type = getattr(self, 'id_type', 'majorbody')
            id_type = self._shared_with['body'].get('id_type', id_type)
            obj = Horizons(id=search_name, id_type=id_type, location='geo', epochs=time.jd)
            eph = obj.ephemerides(extra_precision=True)
            if 'H' in eph.keys():
                self.H = eph['H'][0]
                self.G = eph['G'][0]
            if len(eph['V']) == 1:
                return eph['V'][0]
            else:
                return eph['V'].tolist()

        else:
            obs_obj = self.get_position(time)
            obs_sun = get_sun(time)
            sun_obj = SkyCoord(obs_obj.cartesian - obs_sun.cartesian)
            sun_obj.representation_type = 'spherical'

            # Calculates the phase angle between the 2-vectors
            unit_vector_1 = -obs_obj.cartesian.xyz / np.linalg.norm(obs_obj.cartesian.xyz)
            unit_vector_2 = -sun_obj.cartesian.xyz / np.linalg.norm(sun_obj.cartesian.xyz)
            dot_product = np.dot(unit_vector_1, unit_vector_2)
            phase = np.arccos(dot_product).to(u.deg).value

            return apparent_magnitude(self.H, self.G, obs_obj.distance.to(u.AU).value,
                                      sun_obj.distance.to(u.AU).value, phase)

    @deprecated_function(message="Please use get_pole_position_angle from Body object")
    def get_pole_position_angle(self, pole, time):
        """Returns the object's geocentric position.

        Parameters
        ----------
        pole : `str`, `astropy.coordinates.SkyCoord`
            Coordinate of the object pole ICRS.

        time : `str`, `astropy.time.Time`
            Time from which to calculate the position. It can be a string
            in the ISO format (yyyy-mm-dd hh:mm:ss.s) or an astropy Time object.

        Returns
        -------
        position_angle : `float`
            Position angle of the object pole, in degrees.

        aperture_angle : `float`
            Aperture angle of the object pole, in degrees.
        """
        time = Time(time)
        if type(pole) == str:
            pole = SkyCoord(pole, unit=(u.hourangle, u.deg))
        obj = self.get_position(time)
        position_angle = obj.position_angle(pole).value * u.rad
        aperture_angle = np.arcsin(
            -(np.sin(pole.dec) * np.sin(obj.dec) +
              np.cos(pole.dec) * np.cos(obj.dec) * np.cos(pole.ra - obj.ra))
        )
        return position_angle.to('deg'), aperture_angle.to('deg')

    # End of block removal for v1.0

    def get_ksi_eta(self, time, star):
        """Returns the object's projected position relative to a star.

        Returns the projected position (orthographic projection) of the object
        in the tangent sky plane relative to a star.

        Parameters
        ----------
        time : `str`, `astropy.time.Time`
            Reference time to calculate the object position. It can be a string
            in the ISO format (yyyy-mm-dd hh:mm:ss.s) or an astropy Time object.

        star : `str`, `astropy.coordinates.SkyCoord`
            Coordinate of the star in the same reference frame as the ephemeris.

        Returns
        -------
        ksi, eta : `float`
            Projected position (orthographic projection) of the object in the
            tangent sky plane relative to a star.
            ``ksi`` is in the North-South direction (North positive).
            ``eta`` is in the East-West direction (East positive).
        """
        from astropy.coordinates import SkyOffsetFrame
        time = Time(time)
        if type(star) == str:
            star = SkyCoord(star, unit=(u.hourangle, u.deg))
        coord = self.get_position(time)
        target = coord.transform_to(SkyOffsetFrame(origin=star))
        da = target.cartesian.y
        dd = target.cartesian.z
        return da.to(u.km).value, dd.to(u.km).value

    def add_offset(self, da_cosdec, ddec):
        """Adds an offset to the Ephemeris.

        Parameters
        ----------
        da_cosdec : `int`, `float`
            Delta_alpha_cos_delta, in mas.

        ddec : `int`, `float`
            Delta_delta, in mas.
        """
        self.offset = (da_cosdec, ddec)

    @property
    def offset(self):
        return self._offset

    @offset.setter
    def offset(self, value):
        from astropy.coordinates import SphericalCosLatDifferential
        dadc, dd = value
        self._offset = SphericalCosLatDifferential(dadc * u.mas, dd * u.mas, 0.0 * u.km)

    def to_file(self, time, namefile=None):
        """Saves the ephemerides to a file.

        Ephemeris will be saved starting one hour before the central time
        untill one hour after it, with a step of one minute.

        Note
        ----
            This file can be used as an input for ``EphemPlanete()``.

        Parameters
        ----------
        time : `str`, `astropy.time.Time`
            Central time to be saved.

        namefile : `str`
            Filename to save.
        """
        if namefile is None:
            namefile = 'Ephem_' + self.name.replace(' ', '_') + '.dat'
        time = input_tests.test_attr(time, Time, 'time')
        time_output = time + np.arange(-60, 61, 1) * u.min
        ephem_output = self.get_position(time_output.utc)
        array_output = np.array([time_output.utc.jd, ephem_output.ra.deg, ephem_output.dec.deg,
                                 ephem_output.distance.au]).T
        np.savetxt(namefile, array_output, delimiter='    ', fmt='%.14f')

    def __str__(self):
        """String representation of the Ephem Class.
        """
        out = ("----------- Ephemeris -----------\n"
               "\n{}: {{ephem_info}} (SPKID={})\n"
               "Ephem Error: RA*cosDEC: {:.3f}; DEC: {:.3f}\n".format(
                   self.__class__.__name__, getattr(self, 'spkid', ''), self.error_ra, self.error_dec)
               )
        if hasattr(self, 'offset'):
            out += 'Offset applied: RA*cosDEC: {:.4f}; DEC: {:.4f}\n'.format(
                self.offset.d_lon_coslat.to(u.arcsec),
                self.offset.d_lat.to(u.arcsec))
        return out



import os

import numpy as np

from sora.config.decorators import deprecated_alias

__all__ = ['calc_fresnel', 'calc_magnitude_drop']


@deprecated_alias(lambida='bandpass')  # remove this line for v1.0
def calc_fresnel(distance, bandpass):
    """Calculates the Fresnel scale.

    Fresnel Scale = square root of half the multiplication of wavelength and
    object distance.

    Parameters
    ----------
    distance : `int`, `float` array
        Distances, in km.

    bandpass : `int`, `float`, array
        Wavelength, in km.

    Returns
    -------
    fresnel_scale : `float`, array
        Fresnel scale, in km.
    """
    return np.sqrt(bandpass * distance / 2)


def bar_fresnel(X, X01, X02, fresnel_scale, opacity):
    """Returns the modelled light curve considering fresnel diffraction.

    Parameters
    ----------
    X : array
        Array with time values converted in km using the event velocity.

    X01 : `int`, `float`
        Immersion time converted in km using the event velocity.

    X02 `int`, `float`
        Emersion time converted in km using the event velocity.

    fresnel_scale : `int`, `float`
        Fresnel scale, in km.

    opacity : `int`, `float`
        Opacity. Opaque = 1.0, transparent = 0.0.

    Returns
    -------
    flux_fresnel : array
        The light curve with fresnel diffraction.
    """
    import scipy.special as scsp

    # Converting from km to units of fresnel scale
    x = X / fresnel_scale
    x01 = X01 / fresnel_scale
    x02 = X02 / fresnel_scale
    # Fresnel diffraction parameters
    x1 = x - x01
    x2 = x - x02
    s1, c1 = scsp.fresnel(x1)
    s2, c2 = scsp.fresnel(x2)
    cc = c1 - c2
    ss = s1 - s2
    r_ampli = - (cc + ss) * (opacity / 2.)
    i_ampli = (cc - ss) * (opacity / 2.)
    # Determining the flux considering fresnel diffraction
    flux_fresnel = (1.0 + r_ampli) ** 2 + i_ampli ** 2
    return flux_fresnel


def fit_pol(x, y, deg):
    """Fits a n-degree polynomial to the data.

    Parameters
    ----------
    x : `array`
        x-values.

    y : `array`
        y-values.

    deg : `int`
        Degree of the polynomial.


    Returns
    -------
    param : array
        Array with the fitted values.

    param_err : array
        Array with the errors of the fitted values.
    """
    from scipy.odr import odrpack as odr
    from scipy.odr import models

    func = models.polynomial(deg)
    mydata = odr.Data(x, y)
    myodr = odr.ODR(mydata, func, maxit=200)
    myodr.set_job(fit_type=2)
    fit = myodr.run()
    param = fit.beta[::-1]
    param_err = fit.sd_beta[::-1]
    return param, param_err


def calc_magnitude_drop(mag_star, mag_obj):
    """Determines the magnitude drop of the occultation.

    Parameters
    ----------
    mag_star : `int`, `float`
        Star magnitude.

    mag_obj : `int`, `float`
        Object apparent magnitude to the date.


    Returns
    -------
    mag_drop : `float`
        Magnitude drop for the given magnitudes.

    bottom_flux : `float`
        Normalized bottom flux for the given magnitudes.
    """
    contrast = 1 / (1 + (10 ** ((mag_star - mag_obj) * 0.4)))
    mag_combined = mag_star - (2.5 * (np.log10(1 / contrast)))
    mag_drop = mag_obj - mag_combined
    bottom_flux = 10 ** ((mag_combined - mag_obj) * 0.4)
    mag_drop = mag_drop
    bottom_flux = bottom_flux
    return mag_drop, bottom_flux


def read_lc_file(lc_file, usecols=None, skiprows=0):
    """Reads light curve data from a file.

    Parameters
    ----------
    file : `str`, required
        A file with the time and flux in the first and second columns, respectively.
        A third column with error in flux can also be given.

    usecols : `int`, `tuple`, array, optional
        Which columns to read, with the first being the time, the seconds the
        flux and third the flux error.

    skiprows : `int`, optional, default=0
        Skip the first skiprows lines, including comments.

    Returns
    -------
    :rtype: : (array, array, array)

    """
    from astropy.io import ascii
    from astropy.time import Time

    if not os.path.isfile(lc_file):
        raise ValueError('{} not found'.format(lc_file))
    if usecols is not None:
        if len(usecols) == 2:
            try:
                time, flux = np.loadtxt(lc_file, usecols=usecols, skiprows=skiprows, unpack=True)
                return time, flux
            except:
                pass
        elif len(usecols) == 3:
            try:
                time, flux, dflux = np.loadtxt(lc_file, usecols=usecols, skiprows=skiprows, unpack=True)
                return time, flux, dflux
            except:
                pass
        else:
            raise ValueError('usecols should have 2 or 3 values')
    else:
        try:
            time, flux = np.loadtxt(lc_file, usecols=[0, 1], skiprows=skiprows, unpack=True)
            return time, flux
        except:
            pass
        try:
            time, flux, dflux = np.loadtxt(lc_file, usecols=[0, 1, 2], skiprows=skiprows, unpack=True)
            return time, flux, dflux
        except:
            pass
    try:
        lc_data = ascii.read(lc_file, data_start=skiprows)

        if usecols is not None:
            if len(usecols) == 2:
                time_col_index, flux_col_index = usecols
                time_col_name = lc_data.colnames[time_col_index]
                flux_col_name = lc_data.colnames[flux_col_index]
                time, flux = Time(lc_data[time_col_name].data).jd, lc_data[flux_col_name].data
                return time, flux
            elif len(usecols) == 3:
                time_col_index, flux_col_index, dflux_col_index = usecols
                time_col_name = lc_data.colnames[time_col_index]
                flux_col_name = lc_data.colnames[flux_col_index]
                dflux_col_name = lc_data.colnames[dflux_col_index]
                time, flux = Time(lc_data[time_col_name].data).jd, lc_data[flux_col_name].data
                dflux = lc_data[dflux_col_name].data
                return time, flux, dflux
            else:
                raise ValueError('usecols should have 2 or 3 values')
        else:
            if len(lc_data.colnames) == 2:
                time_col_index, flux_col_index = [0, 1]
                time_col_name = lc_data.colnames[time_col_index]
                flux_col_name = lc_data.colnames[flux_col_index]
                time, flux = Time(lc_data[time_col_name].data).jd, lc_data[flux_col_name].data
                return time, flux
            elif len(lc_data.colnames) == 3:
                time_col_index, flux_col_index, dflux_col_index = [0, 1, 2]
                time_col_name = lc_data.colnames[time_col_index]
                flux_col_name = lc_data.colnames[flux_col_index]
                dflux_col_name = lc_data.colnames[dflux_col_index]
                time, flux = Time(lc_data[time_col_name].data).jd, lc_data[flux_col_name].data
                dflux = lc_data[dflux_col_name].data
                return time, flux, dflux
            else:
                raise ValueError('file columns  should have 2 or 3 values')
    except:
        raise ValueError('An error has occurred in reading the {}'.format(lc_file))



from .body import Body
from .ephem import EphemKernel, EphemPlanete, EphemJPL, EphemHorizons
from .lightcurve import LightCurve
from .observer import Observer, Spacecraft
from .star import Star
from .occultation import Occultation

__version__ = '0.2dev'

print(f'SORA version: {__version__}')
